@@ -5,9 +5,9 @@ import os
 import numpy as np
 from cfl.dataset import Dataset
 from cfl.block import Block
-import cfl.density_estimation_methods as cdem
-import cfl.cluster_methods as ccm
-from cfl import intervention_rec
+import cfl.cond_density_estimation as cdem
+import cfl.clustering as ccm
+from cfl.util.input_val import validate_data_info
 
 '''
 Methods in Experiment Class: 
@@ -28,22 +28,23 @@ Methods in Experiment Class:
 # NOTE: the keys of this dictionary are passed as part of the 'block_names'
 # list. They are different than the names of the attributes in each block's
 # self.name attribute
-BLOCK_KEY = {   'CDE'             : cdem.CDE,
-                'CauseClusterer'  : ccm.CauseClusterer,
-                'EffectClusterer' : ccm.EffectClusterer} #TODO: maybe change this so that instead of 
-                                                            # calling clusterer, 'Kmeans', 'DBSCAN' and 'SNN' are registered as cluster methods  
+BLOCK_KEY = {'CondDensityEstimator': cdem.CondDensityEstimator,
+             'CauseClusterer': ccm.CauseClusterer,
+             'EffectClusterer': ccm.EffectClusterer}  # TODO: maybe change this so that instead of
+# calling clusterer, 'Kmeans', 'DBSCAN' and 'SNN' are registered as cluster methods
+
 
 class Experiment():
     '''The Experiment class: 
- 
+
     - Creates a pipeline to pass data through the different Blocks of CFL 
     - Save parameters, models, results of the pipeline for reuse 
     '''
 
-
-    def __init__(self, data_info, X_train, Y_train, X_train_raw=None, 
-                 Y_train_raw=None, past_exp_path=None, block_names=None, 
-                 block_params=None, blocks=None, verbose=1, results_path=''):
+    def __init__(self, data_info, X_train, Y_train, X_train_raw=None,
+                 Y_train_raw=None, in_sample_idx=None, out_sample_idx=None,
+                 past_exp_path=None, block_names=None,
+                 block_params=None, blocks=None, verbose=1, results_path=None):
         ''' 
         Sets up and trains an Experiment.
 
@@ -63,7 +64,7 @@ class Experiment():
             blocks (list of Blocks): list of block objects. Default is None.
             verbose (int): Amount of output to print. Possible values are 0, 1, 2. Default is 1.
             results_path (str) : path to directory to save this experiment to.
-                Default is `''`.
+                If None, results will not be saved. Default is None.
 
         Note: There are three ways to specify Blocks: 
                 1) specify `past_exp_path`
@@ -73,7 +74,7 @@ class Experiment():
         '''
         self.verbose = verbose
 
-        # OPTION 1 for Experiment initialization: load from path. 
+        # OPTION 1 for Experiment initialization: load from path.
         # if loading from past experiment, make sure no other block
         # specifications are provided
         if past_exp_path is not None:
@@ -83,12 +84,12 @@ class Experiment():
 
             # load in block names and params
             print(f'Loading in Experiment from {past_exp_path}')
-            block_names, block_params = self.__load_params(os.path.join(past_exp_path, 'params'))
+            block_names, block_params = self.__load_params(
+                os.path.join(past_exp_path, 'params'))
 
-
-        # OPTION 2 for Experiment initialization: create blocks from strings. 
-        # make sure block names and params are both provided, and that 
-        # blocks is left unpopulated 
+        # OPTION 2 for Experiment initialization: create blocks from strings.
+        # make sure block names and params are both provided, and that
+        # blocks is left unpopulated
         elif (block_names is not None) or (block_params is not None):
             assert (block_names is not None), 'block_names should be specified.'
             assert (block_params is not None), 'block_params should be specified.'
@@ -97,13 +98,13 @@ class Experiment():
             # add verbosity to params that don't specify
             # removed because was causing problems with some sklearn models that don't have a verbose param
             # block_params = self.__propagate_verbosity(self.verbose, block_params)
-        
-        #OPTION 3 for Experiment initialization: blocks pre-created. 
+
+        # OPTION 3 for Experiment initialization: blocks pre-created.
         # make sure that only blocks is provided.
         elif blocks is not None:
             assert (block_names is None), 'block_names should not be specified.'
             assert (block_params is None), 'block_params should not be specified.'
-            
+
             for block in self.blocks:
                 assert isinstance(block, Block), \
                     'A specified block is not of type Block.'
@@ -117,16 +118,19 @@ class Experiment():
 
         # build and track training dataset
         # Note: explicitly stating one dataset for training as an Experiment
-        # attribute enforces the definition that an Experiment is a unique 
+        # attribute enforces the definition that an Experiment is a unique
         # configuration of a trained CFL.
+        validate_data_info(data_info)
         self.data_info = data_info
         self.datasets = {}
-        self.dataset_train = self.add_dataset(X=X_train, Y=Y_train, \
-                                              dataset_name='dataset_train', \
-                                              Xraw=X_train_raw, \
-                                              Yraw=Y_train_raw)
+        self.dataset_train = self.add_dataset(X=X_train, Y=Y_train,
+                                              dataset_name='dataset_train',
+                                              Xraw=X_train_raw,
+                                              Yraw=Y_train_raw,
+                                              in_sample_idx=in_sample_idx,
+                                              out_sample_idx=out_sample_idx)
         self.datasets[self.dataset_train.get_name()] = self.dataset_train
-        
+
         # build experiment directory
         self.save_path = self.__make_exp_dir(results_path)
 
@@ -134,22 +138,22 @@ class Experiment():
         # or 2)
         if blocks is None:
             blocks = []
-            for bn,bp in zip(block_names, block_params): # data_info
-                blocks.append(self.__build_block(bn,bp))
+            for bn, bp in zip(block_names, block_params):  # data_info
+                blocks.append(self.__build_block(bn, bp))
 
         # load in trained block info if past experiment provided
         if past_exp_path is not None:
             for block in blocks:
-                fn = os.path.join(past_exp_path, 'trained_blocks', block.get_name())
+                fn = os.path.join(
+                    past_exp_path, 'trained_blocks', block.get_name())
                 block.load_block(fn)
             self.is_trained = True
 
-
         # TODO: check that interfaces match
         # TODO: assert in the function itself so we can give more info
         # about what exactly is incompatible
         # assert self.check_blocks_compatibility(), 'Specified blocks are incompatible'
-        
+
         # save configuration parameters for each block
         self.blocks = blocks
         self.block_names = block_names
@@ -170,36 +174,39 @@ class Experiment():
 
         if not self.is_trained:
 
+            if self.verbose > 0:
+                print(20*'#' + ' Beginning CFL Experiment training. ' + 20*'#')
+
             # check inputs
             assert isinstance(dataset, (type(None), Dataset, str)), \
                 'dataset should be None, or of type Dataset or str.'
             assert isinstance(prev_results, (type(None), dict)), \
                 'prev_results should be None or a dict'
 
-            if self.verbose > 0:
-                print('Training CFL pipeline.')
-
             # pull specified dataset
-            if dataset is None: #if you don't specify a dataset, use the one specified in initialization
+            if dataset is None:  # if you don't specify a dataset, use the one specified in initialization
                 dataset = self.get_dataset('dataset_train')
-            elif isinstance(dataset, str): #otherwise, they can pass a string specifying a particular dataset to use
+            # otherwise, they can pass a string specifying a particular dataset to use
+            elif isinstance(dataset, str):
                 if dataset != 'dataset_train':
                     if self.verbose > 0:
-                        print('Warning: you are not using the dataset_train ' + \
-                        'Dataset specified in Experiment initialization for ' + \
-                        'training the CFL pipeline.')
+                        print('Warning: you are not using the dataset_train ' +
+                              'Dataset specified in Experiment initialization for ' +
+                              'training the CFL pipeline.')
                 dataset = self.get_dataset(dataset)
             else:
                 if self.verbose > 0:
-                    print('Warning: by specifying your own Dataset for ' + \
-                        'training, you may not be using the same data as ' + \
-                        'specified for training in Experiment initialization.')
+                    print('Warning: by specifying your own Dataset for ' +
+                          'training, you may not be using the same data as ' +
+                          'specified for training in Experiment initialization.')
 
             all_results = {}
 
-            # this is the main logic - train each block 
+            # this is the main logic - train each block
             for block in self.blocks:
                 # train current block
+                if self.verbose > 0:
+                    print(f'Beginning {block.get_name()} training...')
                 results = block.train(dataset, prev_results)
                 all_results[block.get_name()] = results
 
@@ -207,20 +214,30 @@ class Experiment():
                 self.__save_results(results, dataset, block)
 
                 # save trained block
-                fn = os.path.join(self.save_path, 'trained_blocks', block.get_name())
-                block.save_block(fn)
-                
+                if self.save_path is not None:
+                    fn = os.path.join(self.save_path, 'trained_blocks',
+                                      block.get_name())
+                    block.save_block(fn)
+
                 # pass results on to next block
                 prev_results = results
-            
+
+                if self.verbose > 0:
+                    print(f'{block.get_name()} training complete.')
+
             self.is_trained = True
             dataset.set_cfl_results(all_results)
+
+            if self.verbose > 0:
+                print('Experiment training complete.')
+
             return all_results
-        else: 
-            raise Exception('This Experiment has already been trained. ' + \
-                'If you would like to use a new Dataset for training, ' + \
-                'please create a new Experiment.')
-    
+
+        else:
+            raise Exception('This Experiment has already been trained. ' +
+                            'If you would like to use a new Dataset for training, ' +
+                            'please create a new Experiment.')
+
     def predict(self, dataset, prev_results=None):
         ''' Predict using the trained CFL pipeline. 
 
@@ -233,6 +250,9 @@ class Experiment():
                  (dict of dicts) : dict of results dictionaries from all Blocks.
         '''
 
+        if self.verbose > 0:
+            print('Beginning Experiment prediction.')
+
         # check inputs
         assert isinstance(dataset, (type(None), Dataset, str)), \
             'dataset should be None, or of type Dataset or str.'
@@ -243,12 +263,15 @@ class Experiment():
         if isinstance(dataset, str):
             dataset = self.get_dataset(dataset)
 
-        for bi,block in enumerate(self.blocks):
-            assert block.is_trained, 'Block {} has not been trained yet.'.format(bi)
+        for bi, block in enumerate(self.blocks):
+            assert block.is_trained, 'Block {} has not been trained yet.'.format(
+                bi)
 
         all_results = {}
         for block in self.blocks:
             # predict with current block
+            if self.verbose > 0:
+                print(f'Beginning {block.get_name()} prediction...')
             results = block.predict(dataset, prev_results)
             all_results[block.get_name()] = results
 
@@ -256,14 +279,22 @@ class Experiment():
             self.__save_results(results, dataset, block)
 
             # pass results on to next block
-            prev_results = results    
+            prev_results = results
+
+            if self.verbose > 0:
+                print(f'{block.get_name()} prediction complete.')
 
         dataset.set_cfl_results(all_results)
-        return all_results        
 
-    def add_dataset(self, X, Y, dataset_name, Xraw=None, Yraw=None):
+        if self.verbose > 0:
+            print('Prediction complete.')
+
+        return all_results
+
+    def add_dataset(self, X, Y, dataset_name, Xraw=None, Yraw=None,
+                    in_sample_idx=None, out_sample_idx=None):
         ''' Add a new dataset to be tracked by this Experiment. 
-            
+
             Arguments: 
                 X (np.array) : X data of shape (n_samples, n_x_features) associated with 
                     this Dataset. 
@@ -278,7 +309,7 @@ class Experiment():
             Returns:
                 dataset : the newly constructed Dataset object. (Dataset)
         '''
-        
+
         # check inputs
         assert isinstance(X, np.ndarray), \
             'X should be of type np.ndarray. Actual type: {}'.format(type(X))
@@ -286,24 +317,25 @@ class Experiment():
             'Y should be of type np.ndarray. Actual type: {}'.format(type(Y))
         assert isinstance(Xraw, (np.ndarray, type(None))), \
             'Xraw should be of type np.ndarray or NoneType. ' + \
-                'Actual type: {}'.format(type(Xraw))
+            'Actual type: {}'.format(type(Xraw))
         assert isinstance(Yraw, (np.ndarray, type(None))), \
             'Yraw should be of type np.ndarray or NoneType. ' + \
-                'Actual type: {}'.format(type(Yraw))
+            'Actual type: {}'.format(type(Yraw))
         assert isinstance(dataset_name, str), 'dataset_name should be of ' + \
             'type str. Actual type: {}'.format(type(dataset_name))
 
         # check that we haven't added a Dataset by this name already
         assert dataset_name not in self.datasets.keys(), \
-            'A Dataset named {} has already been added to this Experiment.'.format(dataset_name)
+            'A Dataset named {} has already been added to this Experiment.'.format(
+                dataset_name)
 
         # make new Dataset, add to Experiment's dict of datasets
-        dataset = Dataset(X=X, Y=Y, Xraw=Xraw, Yraw=Yraw, name=dataset_name)
+        dataset = Dataset(X=X, Y=Y, name=dataset_name, Xraw=Xraw, Yraw=Yraw,
+                          in_sample_idx=in_sample_idx, out_sample_idx=out_sample_idx)
         self.datasets[dataset_name] = dataset
         return dataset
 
     def get_dataset(self, dataset_name):
-
         ''' Retrieve a Dataset that has been registered with this Experiment.
 
             Arguments:
@@ -321,7 +353,7 @@ class Experiment():
 
         return self.datasets[dataset_name]
 
-    def get_data_info(self): 
+    def get_data_info(self):
         return self.data_info
 
     # TODO: this function doesn't work right now
@@ -343,6 +375,9 @@ class Experiment():
         '''
 
         # check inputs
+        if self.save_path is None:
+            raise FileNotFoundError('results_path was not specified for this \
+                Experiment, so results were not saved.')
         assert isinstance(dataset_name, str), \
             'dataset_name should be of type str.'
 
@@ -358,10 +393,10 @@ class Experiment():
             file_name = os.path.join(dir_name, block.get_name())
             with open(file_name, 'rb') as f:
                 results[block.get_name()] = pickle.load(f)
-        
+
         return results
 
-    def retrieve_results(self, dataset_name='dataset_train'): 
+    def retrieve_results(self, dataset_name='dataset_train'):
         '''Returns the results from running a given dataset
             through the Experiment pipeline. Default is the training dataset
 
@@ -374,23 +409,23 @@ class Experiment():
                         specifies which Block the results come from. The
                         second key specifies the specific result.
         '''
-        dataset = self.datasets['dataset_train']
+        dataset = self.datasets[dataset_name]
         return dataset.cfl_results
 
     # def check_blocks_compatibility(self):
     #     ''' Check that each Block in self.blocks is compatible with the Blocks
-    #         that precede and follow it. 
+    #         that precede and follow it.
     #         TODO: This is currently not implemented and simply states that all
     #         Blocks are always compatible. We may use class registration here
     #         to determine block compatibility. Alternatively, we may have each
     #         Block descendent specify it's argument expectations and return
-    #         guarantees. 
+    #         guarantees.
 
     #         Arguments: None
-            
+
     #         Returns:
     #             compatible : whether or not self.blocks is a compatible
-    #                          composition of Blocks. (bool) 
+    #                          composition of Blocks. (bool)
     #     '''
 
     #     return True
@@ -402,28 +437,29 @@ class Experiment():
         '''
         return self.save_path
 
-    def get_intervention_recs(self, dataset_name, k_samples=100, eps=0.5):
-        ''' For a given dataset, this function selects a subset of samples that
-            serve as optimal interventions to perform when testing for confounding
-            in CFL's observational partition.
-            TODO: complete documentation
-        '''
-        assert self.blocks is not None, 'blocks have not been defined yet.'
-        
-        cfl_results = self.get_dataset(dataset_name).get_cfl_results()
-        assert cfl_results is not None, 'There are no results for this Dataset.'
+    # TODO: remove this? no longer matches our workflow
+    # def get_intervention_recs(self, dataset_name, k_samples=100, eps=0.5):
+    #     ''' For a given dataset, this function selects a subset of samples that
+    #         serve as optimal interventions to perform when testing for confounding
+    #         in CFL's observational partition.
+    #         TODO: complete documentation
+    #     '''
+    #     assert self.blocks is not None, 'blocks have not been defined yet.'
 
-        # TODO: standardize block names as CDE and Clusterer so those can be
-        # hardcoded here. Right now there's no way to know if blocks[0] is actually
-        # a CDE
-        return intervention_rec.get_recommendations(
-                    pyx=cfl_results[self.blocks[0].get_name()]['pyx'], 
-                    cluster_labels=cfl_results[self.blocks[1].get_name()]['x_lbls'], 
-                    k_samples=k_samples, 
-                    eps=eps,
-                    to_plot=False,
-                    series='series'
-                )    
+    #     cfl_results = self.get_dataset(dataset_name).get_cfl_results()
+    #     assert cfl_results is not None, 'There are no results for this Dataset.'
+
+    #     # TODO: standardize block names as CDE and Clusterer so those can be
+    #     # hardcoded here. Right now there's no way to know if blocks[0] is actually
+    #     # a CDE
+    #     return intervention_rec.get_recommendations(
+    #                 pyx=cfl_results[self.blocks[0].get_name()]['pyx'],
+    #                 cluster_labels=cfl_results[self.blocks[1].get_name()]['x_lbls'],
+    #                 k_samples=k_samples,
+    #                 eps=eps,
+    #                 to_plot=False,
+    #                 series='series'
+    #             )
 
     def __propagate_verbosity(self, verbose, block_params):
         for pi in range(len(block_params)):
@@ -432,10 +468,10 @@ class Experiment():
                 modified_params['verbose'] = verbose
                 block_params[pi] = modified_params
         return block_params
-    
+
     def __make_exp_dir(self, results_path):
         ''' Build directory to which to save this Experiment. 
-            
+
             Arguments: 
                 results_path (str): where to build this Experiment's directory. For
                                example, if `results_path='path/to/dir'`, the
@@ -444,9 +480,12 @@ class Experiment():
             Returns: 
                 save_path (str): path to experiment directory 
                             (`'path/to/dir/experiment000x'` in the example above).
-                            
+
         '''
 
+        if results_path is None:
+            return None
+
         # check inputs
         assert isinstance(results_path, str), \
             'results_path should be of type str.'
@@ -454,7 +493,8 @@ class Experiment():
         # make sure base_path exists, if not make it
         if not os.path.exists(results_path):
             if self.verbose > 0:
-                print("save_path '{}' doesn't exist, creating now.".format(results_path))
+                print("save_path '{}' doesn't exist, creating now.".format(
+                    results_path))
             os.makedirs(results_path)
 
         # create dir for this run
@@ -489,10 +529,10 @@ class Experiment():
             'block_param should be of type dict.'
         assert block_name in BLOCK_KEY.keys(), block_name + ' has not been ' + \
             'added to BLOCK_KEY yet. Please do so before proceeding. Note: ' + \
-            'this is a temporary system until we set up Block registration.' 
+            'this is a temporary system until we set up Block registration.'
 
-        return BLOCK_KEY[block_name](data_info=self.data_info, 
-                                     params=block_param)
+        return BLOCK_KEY[block_name](data_info=self.data_info,
+                                     block_params=block_param)
 
     def __save_results(self, results, dataset, block):
         ''' Save results for a given dataset and block. 
@@ -510,7 +550,8 @@ class Experiment():
 
         # check inputs
         assert isinstance(results, dict), 'results should be a dict.'
-        assert isinstance(dataset, Dataset), 'dataset should be of type Dataset.'
+        assert isinstance(
+            dataset, Dataset), 'dataset should be of type Dataset.'
         assert isinstance(block, Block), \
             'block should be of a type that inherits Block.'
 
@@ -519,10 +560,11 @@ class Experiment():
             if not os.path.exists(dir_name):
                 os.mkdir(dir_name)
 
-            file_name = os.path.join(dir_name, block.get_name() + '_results.pickle')
+            file_name = os.path.join(
+                dir_name, block.get_name() + '_results.pickle')
             with open(file_name, 'wb') as f:
-                pickle.dump(results, f) 
-                # TODO: eventually, we have to be careful about what pickle 
+                pickle.dump(results, f)
+                # TODO: eventually, we have to be careful about what pickle
                 # protocol we use for compatibility across python versions
 
     def __save_params(self):
@@ -532,9 +574,10 @@ class Experiment():
 
         if self.save_path is not None:
             assert self.blocks is not None, 'self.blocks does not exist yet.'
-            assert not os.path.exists(os.path.join(self.save_path, 'params')), 'Params already saved.'
+            assert not os.path.exists(os.path.join(
+                self.save_path, 'params')), 'Params already saved.'
             os.mkdir(os.path.join(self.save_path, 'params'))
-            
+
             block_graph = []
             for block in self.blocks:
                 block_graph.append(block.get_name())
@@ -544,22 +587,22 @@ class Experiment():
 
             fn = os.path.join(self.save_path, 'params', 'block_graph')
             with open(fn, 'wb') as f:
-                pickle.dump(block_graph, f) 
-    
+                pickle.dump(block_graph, f)
+
     def __load_params(self, params_path):
         ''' Helper function to load params from a specified previous
             experiment to be used in this experiment. Primarily used in 
             Experiment initialization. 
-            
+
             Arguments: 
                 params_path (str) : path to where params are saved in previous 
                               Experiment. 
-            
+
             Returns: 
                 list of strs : ordered list of blocks used in previous
                               Experiment. Blocks identified by name (should be
                               the same name that block.get_name() returns). 
-                              
+
                 list of dicts : ordered list of params dictionaries associated 
                                with each block. 
         '''
@@ -577,6 +620,7 @@ class Experiment():
 
 ########### HELPER FUNCTIONS ##################################################
 
+
 def get_next_dirname(path):
     ''' gets the next subdirectory name in numerical order. i.e. if  'path' 
     contains 'run0000' and 'run0001', this will return 'run0002'. 
@@ -586,6 +630,7 @@ def get_next_dirname(path):
         next subdirectory name. 
     '''
     i = 0
-    while os.path.exists(os.path.join(path, 'experiment{}'.format(str(i).zfill(4)))): #zfill(4) creates a 4 digit number
-        i += 1  
+    # zfill(4) creates a 4 digit number
+    while os.path.exists(os.path.join(path, 'experiment{}'.format(str(i).zfill(4)))):
+        i += 1
     return 'experiment{}'.format(str(i).zfill(4))
